from base64 import b64decode

from cryptography.exceptions import InvalidKey
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from ormar.exceptions import NoMatch

from kolombo import conf
from kolombo.console import debug
from kolombo.models import User


async def check_credentials(email: str, password: str, domain: str) -> bool:
    key_material = password.encode("utf-8")
    try:
        user = await User.objects.filter(active=True, email=email, domain=domain).get()
        expected_key = b64decode(user.password)
    except NoMatch:
        debug(f"No active user '{email}' from domain '{domain}' found")
        key_material = b"Password that 100% not matching, I swear"
        expected_key = b"Password that 100% not matching, trust me"

    kdf = Scrypt(conf.SALT, length=32, n=2 ** 16, r=8, p=1)
    try:
        kdf.verify(key_material, expected_key)
        return True
    except InvalidKey:
        return False



from os import path

from docker import from_env  # type: ignore[import]
from typer import Argument, Typer

from kolombo.console import error, info, step
from kolombo.util import build_kolombo_image, execute_as_root

dkim_cli = Typer()


def read_dkim_txt_record(domain: str) -> str:
    with open(f"/etc/kolombo/dkim_keys/{domain}.txt", mode="r") as txt_file:
        txt_file_content = txt_file.read()

    paren_open_idx = txt_file_content.find("(")
    paren_close_idx = txt_file_content.find(")")
    txt_record = txt_file_content[paren_open_idx + 1 : paren_close_idx]  # noqa: E203
    return txt_record.replace('"\n\t  "', "")


@dkim_cli.command("generate")
@execute_as_root
def generate_keys(
    domain: str = Argument(..., help="Domain to generate DKIM keys for"),  # noqa: B008
) -> None:
    client = from_env()
    build_kolombo_image(client, "dkim-gen")

    step(f"Generating DKIM keys for domain: {domain}")
    client.containers.run(
        "kolombo-dkim-gen",
        domain,
        stderr=True,
        auto_remove=True,
        volumes=["/etc/kolombo/dkim_keys:/etc/opendkim/keys"],
    )

    dkim_txt = read_dkim_txt_record(domain)
    info(f"[b]TXT[/] record for [b u]mail._domainkey.{domain}[/] is: {dkim_txt}")


@dkim_cli.command("txt")
@execute_as_root
def get_txt_record(
    domain: str = Argument(  # noqa: B008
        ..., help="Domain to get DKIM TXT record for (if it was generated)"
    ),
) -> None:
    if not path.exists(f"/etc/kolombo/dkim_keys/{domain}.txt"):
        error(f"There is no DKIM TXT record generated for {domain}!")
        exit(1)

    dkim_txt = read_dkim_txt_record(domain)
    info(f"[b]TXT[/] record for [b u]mail._domainkey.{domain}[/] is: {dkim_txt}")



from typing import Iterable


def generate_nginx_config(mx: str, domain: str, secret_key: str) -> str:
    from kolombo import __version__ as version

    return (
        f"### Generated by kolombo v{version}\n\n"
        "# IMAP (mail reading)\n"
        "server {\n"
        f"   server_name {mx};\n"
        f"   ssl_certificate /etc/letsencrypt/live/{mx}/fullchain.pem;\n"
        f"   ssl_certificate_key /etc/letsencrypt/live/{mx}/privkey.pem;\n"
        f"   auth_http_header X-Secret-Key {secret_key};\n"
        f"   auth_http_header X-Domain {domain};\n"
        f"   auth_http kolombo-auth:7089/auth;\n"
        "   listen 993 ssl;\n"
        "   protocol imap;\n"
        "}\n\n"
        "# POP3 (mail reading, use IMAP if possible)\n"
        "server {\n"
        f"   server_name {mx};\n"
        f"   ssl_certificate /etc/letsencrypt/live/{mx}/fullchain.pem;\n"
        f"   ssl_certificate_key /etc/letsencrypt/live/{mx}/privkey.pem;\n"
        f"   auth_http_header X-Secret-Key {secret_key};\n"
        f"   auth_http_header X-Domain {domain};\n"
        f"   auth_http kolombo-auth:7089/auth;\n"
        "   listen 995 ssl;\n"
        "   protocol pop3;\n"
        "}\n\n"
        "# Send-only SMTP\n"
        "server {\n"
        f"   server_name {mx};\n"
        f"   ssl_certificate /etc/letsencrypt/live/{mx}/fullchain.pem;\n"
        f"   ssl_certificate_key /etc/letsencrypt/live/{mx}/privkey.pem;\n"
        f"   auth_http_header X-Secret-Key {secret_key};\n"
        f"   auth_http_header X-Domain {domain};\n"
        f"   auth_http kolombo-auth:7089/auth;\n"
        "   listen 465 ssl;\n"
        "   protocol smtp;\n"
        "}\n"
    )


def generate_virtual_domains(domains: Iterable[str]) -> str:
    return "\n".join(domains)


def generate_virtual_ssl_map(mx_domains: Iterable[str]) -> str:
    return "\n".join(
        f"{mx} /etc/letsencrypt/live/{mx}/privkey.pem "
        f"/etc/letsencrypt/live/{mx}/fullchain.pem"
        for mx in mx_domains
    )


def generate_senders_compose_config(domains: Iterable[str]) -> str:
    compose_file = (
        # fmt: off
        "version: '3.8'\n"
        "services:\n"
        # fmt: on
    )
    for index, domain in enumerate(domains):
        compose_file += (
            f"  kolombo-{domain}-sender:\n"
            "     image: kolombo-sender:latest\n"
            "     tty: true\n"
            "     restart: always\n"
            f"     container_name: kolombo-{domain}-sender\n"
            f"     hostname: {domain}\n"
            "     volumes:\n"
            f"       - /etc/kolombo/dkim_keys/{domain}:/etc/opendkim/keys/{domain}\n"
            "     networks:\n"
            "       kolombo-net:\n"
            f"         ipv4_address: 192.168.79.{10 + index}\n"
            f"     command: {domain}\n\n"
        )

    compose_file += (
        # fmt: off
        "networks:\n"
        "  kolombo-net:\n"
        "    external: true\n"
        # fmt: on
    )
    return compose_file



from typer import Typer

from kolombo.configuration_files import generate_senders_compose_config
from kolombo.console import step
from kolombo.util import (
    async_command,
    execute_as_root,
    needs_database,
    run,
    stop_all_kolombo_services,
    stop_kolombo_service,
)

stop_cli = Typer(invoke_without_command=True)


@stop_cli.command("all")
@execute_as_root
@async_command
@needs_database
async def stop_all() -> None:
    from kolombo.models import Domain

    step("Stopping all Kolombo services")
    stop_all_kolombo_services()

    domains = [domain.actual for domain in await Domain.all_active()]
    # Remove duplicates preserving order
    domains = list(dict.fromkeys(domains))
    senders_compose_config = generate_senders_compose_config(domains)
    project_name = "kolombo_senders"
    file_path = "/etc/kolombo/senders-compose.yml"
    with open(file_path, mode="w") as compose_file:
        compose_file.write(senders_compose_config)

    compose_command = ["docker-compose", "-p", project_name, "-f", file_path, "down"]
    run([*compose_command, "--remove-orphans"])


@stop_cli.command("receiver")
@execute_as_root
def stop_receiver() -> None:
    step("Stopping receiver service")
    stop_kolombo_service("receiver")


@stop_cli.command("auth")
@execute_as_root
def stop_auth() -> None:
    step("Stopping auth service")
    stop_kolombo_service("auth")


@stop_cli.command("nginx")
@execute_as_root
def stop_nginx() -> None:
    step("Stopping nginx service")
    stop_kolombo_service("nginx")


@stop_cli.command("senders")
@execute_as_root
@async_command
@needs_database
async def stop_senders() -> None:
    from kolombo.models import Domain

    domains = [domain.actual for domain in await Domain.all_active()]
    # Remove duplicates
    domains = list(dict.fromkeys(domains))
    senders_compose_config = generate_senders_compose_config(domains)
    project_name = "kolombo_senders"
    file_path = "/etc/kolombo/senders-compose.yml"
    with open(file_path, mode="w") as compose_file:
        compose_file.write(senders_compose_config)

    step("Stopping all running senders")
    compose_command = ["docker-compose", "-p", project_name, "-f", file_path, "down"]
    run([*compose_command, "--remove-orphans"])



from datetime import datetime

from fastapi import Header, Response

from kolombo import conf
from kolombo.auth._resolver import get_ip_for_host
from kolombo.auth.credentials import check_credentials
from kolombo.auth.protocol import AuthError, AuthSuccess
from kolombo.console import debug, error, info, warning

#: We only put SMTP-send, IMAP and POP3 (for GMail. Google, use IMAP in 2021 ffs!)
_protocol_to_port_mapping = {"smtp": 25, "imap": 143, "pop3": 110}


async def auth(
    x_secret_key: str = Header(...),  # noqa: B008
    domain: str = Header(..., alias="X-Domain"),  # noqa: B008
    protocol: str = Header(..., alias="Auth-Protocol"),  # noqa: B008
    auth_method: str = Header(...),  # noqa: B008
    email: str = Header(..., alias="Auth-User"),  # noqa: B008
    password: str = Header(..., alias="Auth-Pass"),  # noqa: B008
    login_attempt: int = Header(..., alias="Auth-Login-Attempt"),  # noqa: B008
    client_ip: str = Header(...),  # noqa: B008
) -> Response:
    """Endpoint used for auth of SMTP/IMAP users."""
    now = datetime.now().isoformat()
    if x_secret_key != conf.NGINX_SECRET_KEY:
        # This MUST NOT happen if everything is set up properly
        error(f"({now}) Not nginx trying to use API")
        return AuthError("Go Away", retry=False)

    # Check for possible usage errors to close connection early
    if protocol not in _protocol_to_port_mapping:
        error(f"({now}) Unsupported protocol: {protocol}")
        return AuthError("Unsupported protocol", retry=False)
    elif auth_method != "plain":
        error(f"({now}) Unsupported auth method: {auth_method}")
        return AuthError("Unsupported auth method", retry=False)

    # Remove newline from credentials
    email = email.rstrip("%0A")
    password = password.rstrip("%0A")
    if not await check_credentials(email, password, domain):
        warning(f"({now}) Failed {protocol} auth as {email} from {client_ip}")
        retry = login_attempt < conf.MAX_ATTEMPTS
        return AuthError("Invalid login or password", retry=retry)

    info(f"({now}) Successful {protocol} auth as {email} from {client_ip}")
    server_host = "kolombo-receiver"
    if protocol == "smtp":
        server_host = f"kolombo-{domain}-sender"

    server_ip = get_ip_for_host(server_host)
    server_port = _protocol_to_port_mapping[protocol]
    debug(f"({now}) Forwarding nginx to {server_ip}:{server_port} ({server_host})")
    return AuthSuccess(server_ip, server_port)



from typing import TYPE_CHECKING, List

from typer import Argument, Typer

from kolombo.configuration_files import (
    generate_nginx_config,
    generate_virtual_domains,
    generate_virtual_ssl_map,
)
from kolombo.console import error, finished, info, print_list, started, step, warning
from kolombo.util import async_command, needs_database

if TYPE_CHECKING:
    from kolombo.models import Domain

domain_cli = Typer()


@domain_cli.command("list")
@async_command
@needs_database
async def list_domains() -> None:
    from kolombo.models import Domain

    all_domains = await Domain.all_active()
    active_pairs = [f"(MX) {domain.mx} -> {domain.actual}" for domain in all_domains]
    info(f"Active domains: {len(active_pairs)}")
    if len(active_pairs) > 0:
        print_list(active_pairs)


def update_virtual_files(active_domains: List["Domain"]) -> None:
    domains = (domain.actual for domain in active_domains)
    virtual_domains = generate_virtual_domains(domains)
    with open("/etc/kolombo/virtual/domains", mode="w") as virtual_domains_file:
        virtual_domains_file.write(f"{virtual_domains}\n")

    mx_domains = (domain.mx for domain in active_domains)
    virtual_ssl_map = generate_virtual_ssl_map(mx_domains)
    with open("/etc/kolombo/virtual/ssl_map", mode="w") as virtual_ssl_map_file:
        virtual_ssl_map_file.write(f"{virtual_ssl_map}\n")


@domain_cli.command("add")
@async_command
@needs_database
async def add_domain(
    domain: str = Argument(  # noqa: B008
        ..., help="Domain that comes after @ in email"
    ),
    mx: str = Argument(None, help="Domain from DNS MX record if exists"),  # noqa: B008
) -> None:
    from kolombo import conf
    from kolombo.models import Domain

    if mx is None:
        mx = domain

    if not domain or not mx:
        error("Arguments MUST NOT be empty strings!")
        exit(1)
    elif await Domain.objects.filter(actual=domain, mx=mx).exists():
        error(f"Pair [code]{mx} -> {domain}[/] already exists!")
        exit(1)

    started(f"Adding [code]{mx} -> {domain}[/] pair")

    step("Adding configuration to [code]mail-enabled[/]")
    nginx_config = generate_nginx_config(mx, domain, conf.NGINX_SECRET_KEY)
    with open(f"/etc/kolombo/mail-enabled/{mx}.conf", mode="w") as nginx_file:
        nginx_file.write(nginx_config)

    step("Saving to database")
    await Domain.objects.create(actual=domain, mx=mx)

    step("Updating virtual files (domains and SSL map)")
    active_domains = await Domain.all_active()
    update_virtual_files(active_domains)

    warning(
        f"Run command [code]kolombo dkim generate {domain}[/] to generate DKIM keys"
    )
    warning("Run command [code]kolombo run[/] to reload Kolombo")

    finished(f"Pair [code]{mx} -> {domain}[/] added!")



from os import path

from typer import Context, Typer

from kolombo import conf
from kolombo.console import error
from kolombo.dkim import dkim_cli
from kolombo.domain import domain_cli
from kolombo.init import init
from kolombo.run import run_cli
from kolombo.stop import stop_cli
from kolombo.user import user_cli

kolombo_cli = Typer(name="kolombo", add_completion=True)
kolombo_cli.command("init")(init)
kolombo_cli.add_typer(domain_cli, name="domain")
kolombo_cli.add_typer(dkim_cli, name="dkim")
kolombo_cli.add_typer(user_cli, name="user")
kolombo_cli.add_typer(run_cli, name="run")
kolombo_cli.add_typer(stop_cli, name="stop")


@kolombo_cli.callback()
def main(ctx: Context) -> None:
    if ctx.invoked_subcommand == "init":
        return

    if not path.exists("/etc/kolombo/kolombo.conf"):
        error("Kolombo is not initialized! Run [code]kolombo init[/] first")
        exit(1)

    conf.read_configuration()



from typing import List

from databases import Database
from ormar import Boolean, Integer, Model, String
from sqlalchemy import MetaData, create_engine  # type: ignore[import]

from kolombo import conf

database = Database(conf.DATABASE_URL)
metadata = MetaData()


async def init_database() -> None:
    await database.connect()
    engine = create_engine(conf.DATABASE_URL)
    metadata.create_all(engine)


class Domain(Model):
    class Meta:
        tablename = "domains"
        metadata = metadata
        database = database

    id: int = Integer(primary_key=True)
    #: Domain that is specified in DNS MX record, like mail.example.org
    mx: str = String(max_length=255)
    #: Domain that is actually used in email, like example.org
    actual: str = String(max_length=255)
    #: Whether this domain is active
    active: bool = Boolean(default=True)

    @classmethod
    async def all_active(cls) -> List["Domain"]:
        return await cls.objects.filter(active=True).all()


class User(Model):
    class Meta:
        tablename = "users"
        metadata = metadata
        database = database

    id: int = Integer(primary_key=True)
    #: Domain that user is associated with (part of email after @)
    domain: str = String(max_length=255)
    #: Email like user@domain.org
    email: str = String(max_length=255, unique=True)
    #: Password will be 32 bytes hash base64 encoded, so 44 symbols (+1 just in case)
    password: str = String(max_length=45)
    #: Whether this user is active
    active: bool = Boolean(default=True)

    @classmethod
    async def all_active(cls) -> List["User"]:
        return await cls.objects.filter(active=True).all()



from fastapi import FastAPI

from kolombo import conf
from kolombo.auth.endpoints import auth
from kolombo.models import init_database

conf.read_configuration()
app = FastAPI(
    title="Kolombo auth API",
    debug=conf.DEBUG,
    on_startup=[init_database],
    # Disable Swagger UI and other API docs
    openapi_url=None,
    docs_url=None,
    redoc_url=None,
)
app.add_api_route("/auth", auth, methods=["GET"])



from typing import Dict

from kolombo.console import enable_debug

#: Whether debug mode is enabled
DEBUG: bool = False
#: Secret key that is sent in X-Secret-Key header from nginx
NGINX_SECRET_KEY: str = "changeme"
#: Maximum number of auth attempts before API closes connection with client
MAX_ATTEMPTS: int = 3
#: Salt used for passwords hashing
SALT: bytes = b"changeme"
#: URL used to connect to database
DATABASE_URL: str = "sqlite:////etc/kolombo/kolombo.sqlite"


def _read_config_file(config_path: str) -> Dict[str, str]:
    """Copied from starlette.config.Config"""
    config: Dict[str, str] = {}
    with open(config_path) as config_file:
        for line in config_file.readlines():
            line = line.strip()
            if "=" in line and not line.startswith("#"):
                key, value = line.split("=", 1)
                key = key.strip()
                value = value.strip().strip("\"'")
                config[key] = value

    return config


def read_configuration() -> None:
    config = _read_config_file("/etc/kolombo/kolombo.conf")

    global DEBUG
    DEBUG = bool(config.get("DEBUG", False))
    if DEBUG:
        enable_debug()

    global NGINX_SECRET_KEY
    NGINX_SECRET_KEY = config.get("NGINX_SECRET_KEY", "changeme")

    global MAX_ATTEMPTS
    MAX_ATTEMPTS = int(config.get("MAX_ATTEMPTS", 3))

    global SALT
    SALT = config.get("SALT", "changeme").encode("utf-8")



import sys
from os import getgid, getuid
from pathlib import Path
from pwd import getpwuid

from typer import confirm, prompt

from kolombo.console import debug, enable_debug, finished, info, started, step
from kolombo.util import run

_virtual_configs = {
    "addresses": "bob@example.com bob@example.com",
    "domains": "example.com",
    "mailbox": "bob@example.com bob@example.com/",
    "ssl_map": (
        "example.com /etc/letsencrypt/live/example.com/privkey.pem "
        "/etc/letsencrypt/live/example.com/fullchain.pem"
    ),
}


def init() -> None:
    from kolombo import __version__ as version

    username = getpwuid(getuid()).pw_name
    started(f"Setting up Kolombo for current user [b]{username}[/]")

    step("Creating /etc/kolombo folder ([u]need root privileges[/])")

    info("Creating /etc/kolombo folder (as root)")
    run(["mkdir", "-p", "-m", "750", "/etc/kolombo"], as_root=True)
    info(f"Changing /etc/kolombo owner to {username} (as root)")
    run(["chown", f"{getuid()}:{getgid()}", "/etc/kolombo"], as_root=True)

    step("Writing configuration to /etc/kolombo/kolombo.conf")
    debug_mode = confirm("Enable debug mode?", default=False, show_default=True)
    if debug_mode:
        enable_debug()

    nginx_secret_key: str = prompt(
        "Enter secret key for communication between NginX and auth API",
        default="changeme",
        show_default=True,
        hide_input=True,
        confirmation_prompt=True,
    )
    max_auth_attempts: int = prompt(
        "Enter maximum auth attempts per one session",
        default="3",
        show_default=True,
        type=int,
    )
    passwords_salt: str = prompt(
        "Enter secret key to be used as salt for passwords hashing",
        default="changeme",
        show_default=True,
        hide_input=True,
        confirmation_prompt=True,
    )
    configuration = (
        f"### Generated by kolombo v{version}\n\n"
        "# Whether debug mode is enabled (0 - disabled, 1 - enabled)\n"
        f"DEBUG={int(debug_mode)}\n"
        "# Secret key that is used to determine that nginx is using API\n"
        f"NGINX_SECRET_KEY={nginx_secret_key}\n"
        "# Maximum auth attempts per one session\n"
        f"MAX_ATTEMPTS={max_auth_attempts}\n"
        "# Salt used for passwords hashing\n"
        f"SALT={passwords_salt}\n"
    )
    with open("/etc/kolombo/kolombo.conf", "w") as config_file:
        config_file.write(configuration)

    step("Populating /etc/kolombo with default folders and files")
    debug("Creating /etc/kolombo folders for volumes")
    folders = ("maildirs", "mail-enabled", "virtual", "dkim_keys")
    for folder in folders:
        Path(f"/etc/kolombo/{folder}").mkdir(mode=0o770, exist_ok=True)

    for file in ("addresses", "domains", "mailbox", "ssl_map"):
        debug(f"Writing default file to /etc/kolombo/virtual/{file}")
        with open(f"/etc/kolombo/virtual/{file}", "w") as virtual_file:
            virtual_file.write(f"# {_virtual_configs[file]}\n")

    step("Installing auto-completion ([u]restart current shell session to use[/])")
    run([sys.argv[0], "--install-completion"])

    finished("Kolombo is set up!")



"""
Responses that API MUST return, according to nginx mail auth protocol:
https://nginx.org/en/docs/mail/ngx_mail_auth_http_module.html#protocol
"""
from fastapi import Response


class AuthSuccess(Response):
    """
    Response that MUST be returned after successful auth.

    Parameters:

    * server_ip - IP that nginx will use to access real SMTP/IMAP server.
    * server_port - port that nginx will use to access real SMTP/IMAP server.
    """

    def __init__(self, server_ip: str, server_port: int) -> None:
        headers = {
            "Auth-Status": "OK",
            "Auth-Server": server_ip,
            "Auth-Port": str(server_port),
        }
        super().__init__(content=None, status_code=200, headers=headers)


class AuthError(Response):
    """
    Response that MUST be returned when auth error occured.

    Parameters:

    * status - status that explains error that happened.
    * retry - *(optional)* whether client can retry (default behaviour)
    or connection will be closed.
    * wait - *(optional)* seconds to wait if client can retry (default - 2).
    * error_code - *(optional)* error code that will be used in SMTP case
    (default - "535 5.7.0", a.k.a. "Authentication Failed").
    """

    def __init__(
        self,
        status: str,
        *,
        retry: bool = True,
        wait: int = 2,
        error_code: str = "535 5.7.0",
    ) -> None:
        if status == "OK":
            raise ValueError("Status MUST NOT be 'OK' in error headers")

        headers = {"Auth-Status": status, "Auth-Error-Code": error_code}
        if retry:
            headers["Auth-Wait"] = str(wait)

        super().__init__(content=None, status_code=200, headers=headers)



from base64 import b64encode
from typing import TYPE_CHECKING, List

from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from typer import Argument, Typer, prompt

from kolombo.console import error, finished, info, print_list, started, step, warning
from kolombo.util import async_command, needs_database

if TYPE_CHECKING:
    from kolombo.models import User

user_cli = Typer()


@user_cli.command("list")
@async_command
@needs_database
async def list_users() -> None:
    from kolombo.models import User

    active_users = [user.email for user in await User.all_active()]
    info(f"Active users: {len(active_users)}")
    if len(active_users) > 0:
        print_list(active_users)


async def _save_user(email: str, password: str, domain: str) -> None:
    from kolombo import conf
    from kolombo.models import User

    kdf = Scrypt(conf.SALT, length=32, n=2 ** 16, r=8, p=1)
    b64_password = b64encode(kdf.derive(password.encode("utf-8")))
    await User.objects.create(email=email, password=b64_password, domain=domain)


def update_virtual_files(active_users: List["User"]) -> None:
    emails = [user.email for user in active_users]
    addresses = "\n".join(f"{email} {email}" for email in emails)
    with open("/etc/kolombo/virtual/addresses", mode="w") as addresses_file:
        addresses_file.write(f"{addresses}\n")

    mailboxes = "\n".join(f"{email} {email}/" for email in emails)
    with open("/etc/kolombo/virtual/mailbox", mode="w") as mailbox_file:
        mailbox_file.write(f"{mailboxes}\n")


@user_cli.command("add")
@async_command
@needs_database
async def add_user(
    email: str = Argument(..., help="Email for new user"),  # noqa: B008
) -> None:
    from kolombo.models import Domain, User

    if "@" not in email:
        error(f"Email '{email}' does not contain '@'!")
        exit(1)

    domain = email.split("@", maxsplit=1)[1].strip()
    if not domain:
        error("Domain part MUST NOT be empty string!")
        exit(1)
    elif not await Domain.objects.filter(active=True, actual=domain).exists():
        error(f"Domain '{domain}' is not added (or inactive)!")
        warning(
            f"You can add it via [code]kolombo domain add {domain} mx.{domain}[/code]"
        )
        exit(1)
    elif await User.objects.filter(email=email).exists():
        error(f"User with email '{email}' already exists!")
        exit(1)

    started(f"Adding [code]{email}[/] user")

    password = prompt(f"{email} password", hide_input=True, confirmation_prompt=True)
    step("Saving to database")
    await _save_user(email, password, domain)

    step("Updating virtual files (addresses and mailbox map)")
    active_users = await User.all_active()
    update_virtual_files(active_users)

    warning("Run command [code]kolombo run[/] to reload Kolombo")

    finished(f"User '{email}' added!")



import sys
from asyncio import run as run_async
from functools import wraps

if sys.version_info >= (3, 9):
    from importlib.resources import files
else:
    from importlib_resources import files

from os import execvp, getuid
from os.path import realpath

# Security implications are considered
from subprocess import STDOUT, check_output  # nosec: B404
from typing import Any, Awaitable, Callable, List

from docker import DockerClient  # type: ignore[import]

from kolombo.console import step, warning


def run(command: List[str], as_root: bool = False) -> str:
    if as_root and getuid() != 0:
        command.insert(0, "sudo")

    # We trust input here, it does not come from user
    return check_output(command, text=True, stderr=STDOUT)  # nosec: B603


def async_command(func: Callable[..., Awaitable[None]]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        run_async(func(*args, **kwargs))

    return wrapper


def needs_database(
    func: Callable[..., Awaitable[None]]
) -> Callable[..., Awaitable[Any]]:
    @wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        from kolombo.models import init_database

        await init_database()
        await func(*args, **kwargs)

    return wrapper


def execute_as_root(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if getuid() == 0:
            func(*args, **kwargs)
            return

        warning("Executing as root via sudo!")
        script_path = realpath(sys.argv[0])
        arguments = sys.argv[1:]
        execvp("sudo", ["sudo", script_path, *arguments])  # nosec: B606, B607

    return wrapper


def build_kolombo_image(client: DockerClient, service: str) -> None:
    kolombo_folder = files("kolombo")
    step(f"Building [code]kolombo-{service}[/] image")
    client.images.build(
        tag=f"kolombo-{service}",
        path=str(kolombo_folder),
        pull=True,
        nocache=True,
        rm=True,
        dockerfile=f"docker/Dockerfile.{service}",
    )


def up_all_kolombo_services() -> None:
    project_name = "kolombo_services"
    docker_folder = files("kolombo") / "docker"
    file_path = str(docker_folder / "services" / "docker-compose.yml")
    compose_command = ["docker-compose", "-p", project_name, "-f", file_path, "up"]
    run([*compose_command, "--force-recreate", "-d"])


def up_kolombo_service(service: str) -> None:
    project_name = "kolombo_services"
    docker_folder = files("kolombo") / "docker"
    file_path = str(docker_folder / "services" / "docker-compose.yml")
    compose_command = ["docker-compose", "-p", project_name, "-f", file_path, "up"]
    run([*compose_command, "--no-deps", "--force-recreate", "-d", f"kolombo-{service}"])


def stop_all_kolombo_services() -> None:
    project_name = "kolombo_services"
    docker_folder = files("kolombo") / "docker"
    file_path = str(docker_folder / "services" / "docker-compose.yml")
    run(["docker-compose", "-p", project_name, "-f", file_path, "down"])


def stop_kolombo_service(service: str) -> None:
    project_name = "kolombo_services"
    docker_folder = files("kolombo") / "docker"
    file_path = str(docker_folder / "services" / "docker-compose.yml")
    compose_command = ["docker-compose", "-p", project_name, "-f", file_path, "rm"]
    run([*compose_command, "--stop", "--force", f"kolombo-{service}"])



from docker import from_env  # type: ignore[import]
from typer import Typer

from kolombo.configuration_files import generate_senders_compose_config
from kolombo.console import error, step
from kolombo.util import (
    async_command,
    build_kolombo_image,
    execute_as_root,
    needs_database,
    run,
    up_all_kolombo_services,
    up_kolombo_service,
)

run_cli = Typer()


@run_cli.command("all")
@execute_as_root
@async_command
@needs_database
async def run_all() -> None:
    from kolombo.models import Domain

    client = from_env()
    build_kolombo_image(client, "receiver")
    build_kolombo_image(client, "auth")
    build_kolombo_image(client, "nginx")
    build_kolombo_image(client, "sender")

    step("Deploying all Kolombo services")
    up_all_kolombo_services()

    domains = [domain.actual for domain in await Domain.all_active()]
    # Remove duplicates preserving order
    domains = list(dict.fromkeys(domains))
    if len(domains) < 1:
        error("No active domains to run senders for")
        return

    senders_compose_config = generate_senders_compose_config(domains)
    project_name = "kolombo_senders"
    file_path = "/etc/kolombo/senders-compose.yml"
    with open(file_path, mode="w") as compose_file:
        compose_file.write(senders_compose_config)

    step(f"Deploying senders for domains: {', '.join(domains)}")
    compose_command = ["docker-compose", "-p", project_name, "-f", file_path, "up"]
    run([*compose_command, "--force-recreate", "--remove-orphans", "-d"])


@run_cli.command("receiver")
@execute_as_root
def run_receiver() -> None:
    client = from_env()
    build_kolombo_image(client, "receiver")

    step("Bringing up receiver service")
    up_kolombo_service("receiver")


@run_cli.command("auth")
@execute_as_root
def run_auth() -> None:
    client = from_env()
    build_kolombo_image(client, "auth")

    step("Bringing up auth service")
    up_kolombo_service("auth")


@run_cli.command("nginx")
@execute_as_root
def run_nginx() -> None:
    client = from_env()
    build_kolombo_image(client, "nginx")

    step("Bringing up nginx service")
    up_kolombo_service("nginx")


@run_cli.command("senders")
@execute_as_root
@async_command
@needs_database
async def run_senders() -> None:
    from kolombo.models import Domain

    domains = [domain.actual for domain in await Domain.all_active()]
    # Remove duplicates preserving order
    domains = list(dict.fromkeys(domains))
    if len(domains) < 1:
        error("No active domains to run senders for")
        exit(1)

    client = from_env()
    build_kolombo_image(client, "sender")

    senders_compose_config = generate_senders_compose_config(domains)
    project_name = "kolombo_senders"
    file_path = "/etc/kolombo/senders-compose.yml"
    with open(file_path, mode="w") as compose_file:
        compose_file.write(senders_compose_config)

    step(f"Deploying senders for domains: {', '.join(domains)}")
    compose_command = ["docker-compose", "-p", project_name, "-f", file_path, "up"]
    run([*compose_command, "--force-recreate", "--remove-orphans", "-d"])



import os
import re

from setuptools import find_packages, setup

# Only Kolombo CLI requirements are installed by default
with open("requirements.txt") as reqs:
    required = reqs.read().splitlines()


def get_version(package):
    """Return package version as listed in `__version__` in `__init__.py`"""
    with open(os.path.join(package, "__init__.py")) as f:
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", f.read()).group(1)


def get_long_description():
    """Return the README"""
    with open("README.md", encoding="utf8") as f:
        return f.read()


setup(
    name="kolombo",
    python_requires=">=3.8",
    version=get_version("kolombo"),
    description="Kolombo - CLI for easy mail server managing 💌",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/HarrySky/kolombo",
    project_urls={
        "Documentation": "https://docs.neigor.me/kolombo",
        "Changelog": "https://github/HarrySky/kolombo/blob/main/README.md",
        "Source": "https://github.com/HarrySky/kolombo",
        "Tracker": "https://github.com/HarrySky/kolombo/issues",
    },
    license="Apache License 2.0",
    author="Igor Nehoroshev",
    author_email="hi@neigor.me",
    maintainer="Igor Nehoroshev",
    maintainer_email="hi@neigor.me",
    packages=find_packages(exclude=["tests"]),
    # Use MANIFEST.in for data files
    include_package_data=True,
    zip_safe=False,
    install_requires=required,
    entry_points={"console_scripts": ["kolombo = kolombo.bin:kolombo_cli"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Communications :: Email",
        "Typing :: Typed",
    ],
)



from typing import Any, List

from rich.console import Console
from rich.markdown import Markdown

_console = Console()
_debug_mode = False


def started(message: str) -> None:
    """Message to indicate that command is started"""
    _console.print(message, style="underline")
    _console.line()


def step(message: str) -> None:
    """Message to indicate current step"""
    _console.print(f"> {message}", style="bold green")


def finished(message: str) -> None:
    """Message to indicate that command is finished"""
    _console.line()
    _console.print(f":checkered_flag: {message}", style="green")


def info(message: str) -> None:
    _console.print(message)


def print_list(list_to_print: List[Any]) -> None:
    _console.print(Markdown(f"*  {'* '.join(list_to_print)}"))


def warning(message: str) -> None:
    _console.print(f":warning-text: {message}", style="underline bold orange1")


def error(message: str) -> None:
    _console.print(f":x-text: {message}", style="bold red")


def enable_debug() -> None:
    global _debug_mode
    _debug_mode = True


def debug(message: str) -> None:
    if _debug_mode:
        _console.print(f"[DEBUG] {message}", style="italic cyan")
